/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#if defined(__LINUX__) 

#include <stdio.h>
#include <string.h>

#include "KCriticalSection.h"

#include "KCanServer.h"


KCanServer::KCanServer()
{
  FrameCount = 0;
  IsInitalized = false;
  RingHead = RingTail = 0;
  RingBufferSize = 10240;
  RingBuffer = new KCanFrame[RingBufferSize];
  memset(RingBuffer, 0, sizeof(KCanFrame)*RingBufferSize);
  Semaphore = new KCriticalSection;
}

KCanServer::~KCanServer()
{
  delete Semaphore;
  delete [] RingBuffer;
}

bool KCanServer::Init(const char * CanDev)
{
  if (IsInitalized)
    return false;
  
  IsInitalized = true;
  
  return KRPiCanDriver::Init(CanDev);
}

void KCanServer::Halt()
{
  Terminate();
  while (!Terminated())
    ;
  
  NCanUtils::SleepMs(1000);
}


int  KCanServer::RingCount() const
{
  return (RingHead - RingTail + RingBufferSize) % RingBufferSize;
}

bool KCanServer::RingBufferFull() const
{
  return RingCount() > 3*(RingBufferSize / 4);
}

void KCanServer::IncRingPtr(volatile int & RingPtr)
{
  RingPtr = (RingPtr + RingBufferSize + 1) % RingBufferSize;
}

void KCanServer::PutFrame(const KCanFrame & Frame)
{
  if (Semaphore->Acquire())
  {
    IncRingPtr(RingHead);
    if (RingTail == RingHead)
      IncRingPtr(RingTail);
    RingBuffer[RingHead] = Frame;
    
    Semaphore->Release();
  }
}

bool KCanServer::GetFrame(KCanFrame & Frame)
{
  bool Ok = false;
  if (RingHead == RingTail)
    return Ok;
  
  if (Semaphore->Acquire())
  {
    if (RingTail != RingHead)
    {
      Ok = true;
      IncRingPtr(RingTail);
      Frame = RingBuffer[RingTail];
    }
    Semaphore->Release();
  }
  
  return Ok;
}

void KCanServer::Execute()
{
#if defined(__CONSOLE__) && defined(__DEBUG__)
  printf("start CanServer\n");
#endif
  Init(NULL);
  if (!Connect())
    Terminate();
    
  while (!Terminated())
  {
    KCanFrame Frame;
    
    if (ReceiveData(Frame))
    {
      Frame.Counter = (int)FrameCount++;
      PutFrame(Frame);
    }
  }
#if defined(__CONSOLE__) && defined(__DEBUG__)
  printf("stop CanServer\n");
#endif
}

#endif

