/*
 *
 *  Copyright (C) 2014 Jrg Mller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <stdio.h>
#include <string.h>

#include "KXmlNodeList.h"


void KXmlNodeList::SetNodeValue(const char * Value, int ValueLen)
{
  if (mNodeValue)
    delete [] mNodeValue;
  mNodeValue = NULL;

  if (Value && ValueLen > 0)
  {
    mNodeValue = new char [ValueLen+1];
    memcpy(const_cast<char *> (mNodeValue), Value, ValueLen);
    const_cast<char *> (mNodeValue)[ValueLen] = 0;
  }
}

////////////////////////////////////////////////////////////////////////////////
//
// KXmlNodeList
//
////////////////////////////////////////////////////////////////////////////////

void KXmlNodeList::AppendAttr(const char * Name, const char * ValuePos, int ValueLen)
{
  KXmlAttrList * Attr = new KXmlAttrList;

  if (Name && Name[0])
  {
    Attr->mName = new char[strlen(Name)+1];
    strcpy(const_cast<char *>(Attr->mName), Name);
  }
  if (ValuePos && ValueLen > 0)
  {
    char buffer[1024];
    
    if (ValueLen >= 1023)
      ValueLen = 1023;
    
    strncpy(buffer, ValuePos, ValueLen);
    unsigned len_out = 0;
    for (int i = 0; i < ValueLen; i++)
    {
      char c = buffer[i];
      if (c == '&')
      {
        if (!strncmp(buffer + i, "&amp;", 5))
        {
          c = '&';
          i += 4;
        } else
        if (!strncmp(buffer + i, "&apos;", 6))
        {
          c = '\'';
          i += 5;
        } else
        if (!strncmp(buffer + i, "&quot;", 6))
        {
          c = '"';
          i += 5;
        }
      }
      buffer[len_out++] = c;
    }
    buffer[len_out] = 0;
 
    Attr->mValue = new char[len_out + 1];
    strcpy(const_cast<char *>(Attr->mValue), buffer);
  }
  if (!mAttrList)
    mAttrList = Attr;
  else {
    KXmlAttrList * AttrList = mAttrList;
    while (AttrList->mNext)
      AttrList = AttrList->mNext;
    AttrList->mNext = Attr;
  }
}

const char * KXmlNodeList::FindAttr(const char * Name) const
{
  const KXmlAttrList * AttrList = mAttrList;
  while (AttrList)
  {
    if (AttrList->mName &&
        !strcmp(AttrList->mName, Name))
      return AttrList->mValue;

    AttrList = AttrList->mNext;
  }
  return NULL;
}

const char * KXmlNodeList::FindNodeAttr(const char * NodeName, const char * Name)
{
  const KXmlNode * Node = Find(NodeName);
  
  return Node ? Node->FindAttr(Name) : NULL;
}

KXmlNodeList::KXmlNodeList()
{
  mName[0] = 0;
  mFirstNode = NULL;
  mAttrList = NULL;
  mNodeValue = NULL;
}

KXmlNodeList::~KXmlNodeList()
{
  Init();
}

void KXmlNodeList::Init()
{
  mName[0] = 0;
  
  while (mFirstNode)
  {
    KXmlNode * Node = mFirstNode->mNext;

    delete mFirstNode;

    mFirstNode = Node;
  }

  if (mAttrList)
    delete mAttrList;
  mAttrList = NULL;

  SetNodeValue(NULL, 0);
}

void KXmlNodeList::Append(KXmlNode * Node)
{
  Node->mNext = NULL;
  if (mFirstNode)
  {
    KXmlNode * node = mFirstNode;

    while (node->mNext)
      node = node->mNext;

    node->mNext = Node;
  } else
    mFirstNode = Node;
}

const KXmlAttrList * KXmlNodeList::GetPrevAttr(const KXmlAttrList * Attr) const
{
  if (Attr == mAttrList)
    return NULL;
    
  const KXmlAttrList * ResAttr = mAttrList;

  while (ResAttr && ResAttr->mNext != Attr)
    ResAttr = ResAttr->mNext;

  return ResAttr;
}

bool KXmlNodeList::GetNextNode(const KXmlNode * & Node) const
{
  if (!Node)
    Node = mFirstNode;
  else
    Node = Node->mNext;

  return Node != NULL;
}

bool KXmlNodeList::GetPrevNode(const KXmlNode * & Node) const
{
  if (Node == mFirstNode)
  {
    Node = NULL;
    return false;
  }
  const KXmlNode * ResNode = mFirstNode;
  while (ResNode && ResNode->mNext != Node)
    ResNode = ResNode->mNext;

  Node = ResNode;

  return Node != NULL;
}

KXmlNode * KXmlNodeList::GetLastNode() const
{
  const KXmlNode * ResNode = NULL;

  GetPrevNode(ResNode);

  return const_cast<KXmlNode *> (ResNode);
}
 
KXmlNode * KXmlNodeList::FindNext(const char * Name, KXmlNode * NextNode)
{
  if (!Name || !Name[0])
    return NULL;

  while (NextNode)
  {
    if (NextNode->mName &&
        !strcmp(NextNode->mName, Name))
      return NextNode;

    NextNode = NextNode->mNext;
  }
  return NULL;
}

KXmlNode * KXmlNodeList::Find(const char * Name)
{
  return FindNext(Name, mFirstNode);
}


////////////////////////////////////////////////////////////////////////////////
//
// KXmlAttrList
//
////////////////////////////////////////////////////////////////////////////////

KXmlAttrList::~KXmlAttrList()
{
  if (mName)
    delete [] mName;
  if (mValue)
    delete [] mValue;
  if (mNext)
    delete mNext;
}



