/*
 *
 *  Copyright (C) 2014 Jrg Mller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <vcl.h>
#pragma hdrstop

#include "UCS_Dlg.h"
#include "Comm.hpp"
#include "KTcpServer.cpp"
#include "NCanUtils.h"
#include "KElsterTable.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;

bool Stop = false;
TComm * CS_Com = NULL;
TComm * SecCom = NULL;
KTcpCanClient CanClient;

enum {dev_nullmodem, dev_can_server, dev_can232 } dev;

unsigned char not_connected_dev[256];

bool ElsterComm(const KComfortFrame & send, KComfortFrame & recv);
unsigned short GetKennwort();

__fastcall TForm1::TForm1(TComponent* Owner)
  : TForm(Owner)
{
}

void __fastcall TForm1::FormCreate(TObject *Sender)
{
  char file_name[1024];

  cbDevice->ItemIndex = 1;
  dev = dev_can_server;
  pcMain->ActivePage = tsStart;
  edLog->Text = ExtractFilePath(Application->ExeName) + "cs_log.txt";
  
  int h = cbDevice->Top - cbComm->Top;
  lbRaspi->Top = lbDevice->Top + h;
  lbSecondCom->Top = lbDevice->Top + h;
  edRaspi->Top = cbDevice->Top + h;
  cbSecondCom->Top = cbDevice->Top + h;

  cbDeviceChange(NULL);

  Types::TPoint Size;
  Size.x = 80;
  Size.y = 2000;
  CS_Com = new TComm(this);
  CS_Com->NewScreenSize(Size);
  CS_Com->DTR = true;
  CS_Com->RTS = true;
  CS_Com->Parity = 2;
  CS_Com->StopBits = 2;
  CS_Com->Baudrate = 2400;
  CS_Com->RtsFlowControl = false;
  CS_Com->TimerInterval = 100;
}

void TForm1::EnableControls(bool Ok)
{
  cbComm->Enabled = Ok;
  cbDevice->Enabled = Ok;
  edRaspi->Enabled = Ok;
  cbSecondCom->Enabled = Ok;
  cbxLog->Enabled = Ok;
  btnFileOpen->Enabled = Ok;
  edLog->Enabled = Ok;
}

void __fastcall TForm1::cbDeviceChange(TObject *Sender)
{
  if (0 <= cbDevice->ItemIndex && cbDevice->ItemIndex < 3)
  {
    bool UseCom = cbDevice->ItemIndex == 0 ||
                  cbDevice->ItemIndex == 2;
    lbRaspi->Visible = !UseCom;
    edRaspi->Visible = !UseCom;
    lbSecondCom->Visible = UseCom;
    cbSecondCom->Visible = UseCom;
    if (UseCom)
    {
      if (cbDevice->ItemIndex == 0)
        lbSecondCom->Caption = "2. Schnittstelle";
      else
        lbSecondCom->Caption = "can232";
    }
    dev = cbDevice->ItemIndex;
  }
}

typedef enum
{
  mt_init,
  mt_get,
  mt_change,
  mt_invalid,
} msg_type;

msg_type GetMsgType(bool filter, const KComfortFrame & cf)
{
  if (cf.IsInitFrame())
    return mt_init;

  if (cf.Data[0] != 0x0d)
    return mt_invalid;

  // Vorsichtsmassnahmen:
  // - als Sender nur die Id 680 zulassen
  // - nur reine Abfragen zulassen
  // - Abfragen ohne 0xfa an das Gert mit ID = 000 kann Nebenwirkungen
  //   auslsen! (deshalb nicht zulassen)
  if (filter)
  {
    if (cf.Data[0] == 0x01)
      return mt_get;
    if (cf.Data[0] == 0x00)
      return mt_change;
  }

  if (cf.Data[3] == 0x01 && (cf.Data[4] & 0xf0) == 0 && // reine Abfragen
        (cf.Data[5] == 0xfa || cf.Data[2] != 0))
    return mt_get;

  // Wert ndern
  // 0d 00 03-00-00 fa ...  (marc2012)
  if (cf.Data[3] == 0x00 && (cf.Data[4] & 0xf0) == 0 &&
      cf.Data[5] == 0xfa && cf.Data[2] != 0)
    return mt_change;

  return mt_invalid;
}

bool ConnectedIndex(const KComfortFrame & cs, unsigned char & Index)
{
  KCanFrame Frame;

  bool Ok = cs.SetCanFrame(Frame);
  if (Ok)
    Ok = (Frame.Data[0] & 0x0f) == 1 &&
         Frame.Index() == 0x000b;

  Index = (unsigned char)((Frame.Data[0] & 0xf0) + (Frame.Data[1] & 0x0f));

  return Ok;
}

bool NotConnected(const KComfortFrame & cs)
{
  unsigned char Index;
  bool Ok = ConnectedIndex(cs, Index);
  
  if (Ok &&
      not_connected_dev[Index] > 1)
    return true;

  return false;
}

void __fastcall TForm1::btnStartClick(TObject *Sender)
{  
  if (btnStart->Caption == "Stopp")
  {
    Stop = true;
    return;
  }
  
  if (!InitCom())
    return;

  unsigned short Kennwort = 0;
  if (dev != dev_nullmodem)
  {
    Kennwort = GetKennwort();
    if (Kennwort)
    {
      Kennwort = (Kennwort << 8) + (Kennwort >> 8);
      printf("Kennwort: %d\n", Kennwort);
    } else {
      printf("ungltiges Kennwort\n");
      if (dev != dev_can_server)
        SecCom->ExitCOM();
      return;
    }
  }
  Stop = false;
  CS_Com->SetComPortNr(cbComm->ItemIndex + 1);
  try
  {
    CS_Com->InitCOM();
  }
  catch (...)
  {
 //   on E: Exception do
 //     Application.MessageBox(PChar(E.Message), 'Fehler der seriellen Schnittstelle');
    Application->MessageBox("Fehler der seriellen Schnittstelle", "Fehler");
  }
  if (!CS_Com->COM_INITIALISIERT())
  {
    CS_Com->Show();
    CS_Com->write(" serielle Schnittstelle konnte nicht geffnet werden!  ");
    if (dev != dev_can_server)
      SecCom->ExitCOM();
    return;
  }

  btnStart->Caption = "Stopp";
  EnableControls(false);

  while (CS_Com->ReadFromCOM() != NULL_CHAR)   // Puffer leeren
    NCanUtils::SleepMs(1);

  if (dev != dev_can_server)
    while (SecCom->ReadFromCOM() != NULL_CHAR)
      NCanUtils::SleepMs(1);
     
  KComfortFrame cs;
  KComfortFrame cs_recv;
  bool written = false;
  int len_cs = 0;
  do
  {
    CS_Com->CommMonitor(NULL);
    if (dev == dev_nullmodem)
    {
      SecCom->CommMonitor(NULL);
      int res2 = SecCom->ReadFromCOM();
      if (res2 < 0x100)
        CS_Com->WriteToCOM((char)(res2));
    }
    int res = CS_Com->ReadFromCOM();
    if (res < 0x100)
    {
      if (dev == dev_nullmodem)
        SecCom->WriteToCOM((char)(res));
        
      cs.Data[len_cs++] = char(res);
      if (cs.Data[0] != 0x0d)
        len_cs = 0;

      if (len_cs == 12)
      {
        KCanFrame Frame;
        char str[64];
        msg_type cf_type;

        if (!cs.SetCanFrame(Frame))
          cf_type = mt_invalid;
        else
          cf_type = GetMsgType(dev == dev_nullmodem, cs);

        if (cf_type != mt_invalid)
        {
          if (cbxCanTelegram->Checked ||
              (cf_type == mt_change && !cbNotChange->Checked))
          {
            NCanUtils::FormatFrame(Frame, false, str);
            printf("send: %s", str);
          }
          if (cbxCS_Telegram->Checked)
          {
            cs.SetToString(str);
            printf("send: %s\n", str);
          }
        }
        if (cf_type == mt_change && cbNotChange->Checked)
        {
          if (FormElsterTable(Frame, str))
            printf("Elster %3.3x.%4.4x.%4.4x: %s\n", Frame.Id, Frame.Index(), Frame.Value(),str);
        } else
        if (cf_type == mt_invalid)
        {
          cs.SetToString(str);
          printf("invalid: %s\n", str);
          cs_recv.SetOk(false);
          if (dev != dev_nullmodem)
            CS_Com->WriteBufferToCOM(cs_recv.Data, 12);
        } else
        if (cf_type == mt_init &&
            cbDevice->ItemIndex != 0)
        {
          cs_recv.SetOkValue(Kennwort);
          if (dev != dev_nullmodem)
            CS_Com->WriteBufferToCOM(cs_recv.Data, 12);
        } else
        if (dev != dev_nullmodem &&
            cf_type == mt_get &&
            NotConnected(cs))
        {
          cs_recv.SetOk(false);
          if (dev != dev_nullmodem)
            CS_Com->WriteBufferToCOM(cs_recv.Data, 12);
        } else
        if (dev == dev_nullmodem)
        {
        } else
        if (ElsterComm(cs, cs_recv))
        {
          bool show_recv = false;
          bool recv_ok = cs_recv.SetCanFrame(Frame);
          if (cbxUseElsterTable->Checked || (cf_type == mt_change && !cbNotChange->Checked))
          {
            if (recv_ok)
              recv_ok = FormElsterTable(Frame, str);
            if (recv_ok)
              printf("Elster %3.3x.%4.4x.%4.4x: %s\n", Frame.Id, Frame.Index(), Frame.Value(), str);
            else
            if (cs_recv.Data[0] != 0xaa && cs_recv.Data[0] != 0x55 &&
                !cbxRecvTelegram->Checked && cf_type != mt_init)
              show_recv = true;
          }
          if (cbxRecvTelegram->Checked || cf_type == mt_init || show_recv)
          {
            if (recv_ok &&cf_type != mt_init)
              NCanUtils::FormatFrame(Frame, false, str);
            else {
              cs_recv.SetToString(str);
              strcat(str, "\n");
            }
            if (cf_type == mt_init)
              printf("init: %s", str);
            else
              printf("recv: %s", str);
          }
          unsigned char con_ind;
          if (ConnectedIndex(cs, con_ind))
          {
            if (cs_recv.Data[0] == 0xaa)
              not_connected_dev[con_ind]++;
          }
          if (dev != dev_nullmodem)
            CS_Com->WriteBufferToCOM(cs_recv.Data, 12);
        }
        len_cs = 0;
        written = true;
      }
      if (written)
      {
        Application->ProcessMessages();
        if (dev != dev_nullmodem)
          CS_Com->writeln();
        written = false;
      }
    } else
    if (res != NULL_CHAR)
    {
      len_cs = 0;
      CS_Com->WriteChar(13);
      CS_Com->WriteChar(10);
    } else
      NCanUtils::SleepMs(1);
    Application->ProcessMessages();
  } while (CS_Com->COM_INITIALISIERT() && !Stop);

  CS_Com->ExitCOM();

  btnStart->Caption = "Start";
  EnableControls(true);
}

bool ElsterComm(const KComfortFrame & send, KComfortFrame & recv)
{
  bool Ok = false;

  try
  {
    switch (dev)
    {
      case dev_nullmodem:
        break;

      case dev_can_server:
        Ok = CanClient.SendComfortFrame(send, recv);
        break;

      case dev_can232:
        break;

      default:
        break;
    }
  }
  catch (...)
  {
   // Application->MessageBox(PChar(E.Message), 'Fehler');

    Application->MessageBox("bertragungsfehler", "Fehler");
    Ok = false;
    if (dev == dev_can_server)
      CanClient.Disconnect();
    CS_Com->ExitCOM();
  }
  return Ok;
}

int GetElsterValue(unsigned short receiver_id, unsigned short elster_index)
{
  KComfortFrame send;
  KComfortFrame recv;
  KCanFrame Frame;

  Frame.InitElsterFrame(0x680, receiver_id, elster_index);
  Frame.Len = 7;
  send.GetCanFrame(Frame);
  bool Ok = ElsterComm(send, recv);
  if (Ok)
    Ok = recv.SetCanFrame(Frame);

  if (Ok)
    return Frame.Value();
  else
    return -1;
}

unsigned short GetKennwort()
{
  int res = GetElsterValue(0x601, 0x000b);
  if (res > 0)
    return (unsigned short) res;

  res = GetElsterValue(0x301, 0x000b);
  if (res > 0)
    return (unsigned short) res;

  // http://www.ip-symcon.de/forum/threads/15888-Wolf-CSZ-mit-ISM1
  // eventuell mit direkter RS232-Schnittststelle an WP
  res = GetElsterValue(0x0000, 0x015b); // GERAETEKONFIGURATION_2
  if (res > 0)
    return (unsigned short) res;

  return 0;
}

bool TForm1::InitCom()
{
  bool Ok = true;

  dev = cbDevice->ItemIndex;

  if (dev == dev_can232)
  {
    printf("Kommunikation fuer CAN232 ist noch nicht implementiert\n");
    return false;
  }

  if (dev == dev_can_server)
  {
    CanClient.SetRemoteHost(edRaspi->Text.c_str());
    CanClient.SetRemotePort(5524);
  } else {
    if (!SecCom)
    {
      Types::TPoint Size;
      Size.x = 80;
      Size.y = 2000;
      SecCom = new TComm(this);
      SecCom->NewScreenSize(Size);
      SecCom->DTR = true;
      SecCom->RTS = true;
      SecCom->Parity = 2;
      SecCom->StopBits = 2;
      SecCom->Baudrate = 2400;
      SecCom->RtsFlowControl = false;
      SecCom->TimerInterval = 100;
    }
    try
    {
      SecCom->SetComPortNr(cbSecondCom->ItemIndex + 1);
      Ok = SecCom->InitCOM();
    }
    catch (...)
    {
      Ok = false;
    }
  }
  return Ok;
}

void __fastcall TForm1::FormCloseQuery(TObject *Sender, bool &CanClose)
{
  CanClose = cbComm->Enabled;
}

void __fastcall TForm1::btnFileOpenClick(TObject *Sender)
{
  OpenDialog->FileName = edLog->Text;
  if (OpenDialog->Execute())
  {
    edLog->Text = OpenDialog->FileName;
  }
}

