unit UTcpCanClient;

interface

uses
  SysUtils, ScktComp;

type

  TTcpCanClient = class(TClientSocket)
  private
    RemotePath : string;
    ContentLength : integer;

  public
    Trace : boolean;
    mConnected : boolean;
    
    function SendBuffer(aBuffer : string; TimeOut : integer = 1000) : boolean;
    function WriteLine(aBuffer : string) : boolean;
    function ReadLine(var aString : string) : boolean;
    function ReadHttpStream(var aStream : string; const aGetString : string) : boolean;
    function ReadElster(var Buffer : string) : boolean;
    function SendElster(Buffer : string; timeout : integer) : boolean;

    function Connect(aTimeout : integer = 10) : boolean;
    procedure Disconnect;

  end;

implementation

const SendTimeOut = 1000;
const ReadTimeOut = 1000;

function TTcpCanClient.SendBuffer(aBuffer : string; TimeOut : integer) : boolean;
begin
  Result := true;
  if aBuffer <> '' then
    Result := Socket.SendBuf(aBuffer[1], Length(aBuffer)) = Length(aBuffer);
end;

function TTcpCanClient.WriteLine(aBuffer : string) : boolean;
begin
  aBuffer := aBuffer + #10#13;
  Result := SendBuffer(aBuffer, SendTimeOut);
end;

function TTcpCanClient.ReadLine(var aString : string) : boolean;
var
  Len : integer;
  Timeout : integer;
  Temp : array [0..3] of char;
begin
  aString := '';
  Len := 0;
  result := false;
  Timeout := ReadTimeOut;
  
  while not Result and
        (Socket.ReceiveBuf(Temp, 1) = 1) do begin
    if (Temp[0] = #13) then begin
    end else
    if (Temp[0] = #10) then begin
      Result := true;
    end else
    if aString <> '' then begin
      aString := aString + Temp[0];
    end;
    Timeout := 100;
  end;
  aString := trim(aString);
end;

procedure TTcpCanClient.Disconnect;
begin
  if mConnected then begin
    Active := false;
    mConnected := false;
  end;

  Close();
end;

function TTcpCanClient.Connect(aTimeout : integer) : boolean;
begin
  ClientType := ctBlocking;
 // Block := true;
  Active := true;
{  while not Active and (aTimeout > 0) do begin
    Sleep(10);
    dec(aTimeout, 1);
  end;  }
  if Active and not mConnected then begin
    mConnected := true;
  end;
  Result := mConnected;
end;

// Anfrage vom Internet Explorer:
//
// GET /path HTTP/1.1              << Unterverzeichnis hier einfügen
// Accept: */*
// Accept-Language: de-ch
// Accept-Encoding: gzip, deflate
// User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; ...
// Host: localhost:8080
// Connection: Keep-Alive
// leere Zeile

function TTcpCanClient.ReadHttpStream(var aStream : string;
  const aGetString : string) : boolean;
var
  HeadOk : boolean;
  Get : string;
  remote : string;
  ParsePtr : PChar;
  Buf : array [0..1023] of char;
begin
  Result := false;
  HeadOk := true;

  ContentLength := 0;
  if Connect(1000) then begin
    if aGetString <> '' then
      Get := aGetString
    else
      Get := 'GET ' +
             RemotePath +
             ' HTTP/1.1'#13;
    remote := 'Host: ' + Socket.RemoteHost;
    Result := WriteLine(Get);
    if Result then
      Result := WriteLine(remote);
    if Result then
      Result := WriteLine('');
    if Result then begin
      Result := ReadLine(aStream);
      if copy(aStream, 1, 14) <> 'HTTP/1.1 200 OK' then
        HeadOk := false;
  //    if (Trace)
  //      printf("Tcp line: %s\n", aStream);
    end;
    if Result then begin
      // lesen bis zur ersten Leerzeile
      // (Header wird mit einer Leerzeile abgeschlossen)
      repeat
        Result := ReadLine(aStream);
        // "Content-Length: %d"
        if Result then begin
   //       if (Trace)
   //         printf("Tcp line: %s\n", aStream);
          ParsePtr := AnsiStrPos(PChar(aStream), 'Content-Length');
          if ParsePtr <> nil then begin
            inc(ParsePtr, Length('Content-Length'));
            while ParsePtr^ = ' ' do
              inc(ParsePtr);
            if ParsePtr = ':' then begin
              inc(ParsePtr);
              ContentLength := StrToIntDef(ParsePtr, 0);
            end;
          end;
        end;
      until not Result or (Length(aStream) = 0);

      if (0 < ContentLength) and (ContentLength < High(Buf)) then begin
        Result := (Socket.ReceiveBuf(Buf, ContentLength) = ContentLength);
        if Result then begin
          SetLength(aStream, ContentLength);
          Move(Buf, aStream[1], ContentLength);
        end else
          aStream := '';
      end else
        Result := ReadLine(aStream);
    end;

    Disconnect();
  end;

  if not HeadOk then
    Result := false;
end;

function TTcpCanClient.ReadElster(var Buffer : string) : boolean;
var
  Read : integer;
  Tot : integer;
  Buf : array [0..63] of char;
  debug : string;
begin
  Result := false;
  Buffer := '';
  Debug := '';
  Read := Socket.ReceiveBuf(Buf, 1); // timeout: 500 ms
  if (Read = 1) and (Buf[0] <> #10) then begin  // auf LF synchronisieren
    for Tot := Read to 3*12 do begin
      Read := Socket.ReceiveBuf(Buf[Tot], 1);
      if Read <> 1 then begin
  //      printf("broken %s\n", Buffer);
        break;
      end;
      Debug := Debug + Buf[Tot];
      if Buf[Tot] = #10 then begin
        SetLength(Buffer, Tot);
        Move(Buf, Buffer[1], Tot);
        Result := true;
        exit;
      end;
    end;
  //  printf("no LF %s\n", Buffer);
  end;
  Debug := Debug + '';
end;

function TTcpCanClient.SendElster(Buffer : string; timeout : integer) : boolean;
begin
  Buffer := trim(Buffer) + #10;

  result := SendBuffer(Buffer, timeout);
end;

end.
 