unit Ulogger;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Comm, StdCtrls, ExtCtrls;

type
  TfrmLogger = class(TForm)
    Edit1: TEdit;
    Label1: TLabel;
    btnStart: TButton;
    cbxComm1: TComboBox;
    cbxComm2: TComboBox;
    btnStop: TButton;
    Timer1: TTimer;
    procedure btnStartClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    { Private declarations }
  public
    log_file : System.Text;
    Comm1 : TComm;
    Comm2 : TComm;
    written : integer;
    flush_delay : integer;
  published
    property Text;
  end;

var
  frmLogger: TfrmLogger;

implementation

{$R *.dfm}

const UseModem : boolean = false;

procedure InitComm(var Comm : TComm);
var
  Size : TPoint;
begin
  Size.X := 80;
  Size.Y := 2000;
  with Comm do begin
    NewScreenSize(Size);
    DTR := true;
    RTS := true;
    if not UseModem then begin
      Parity := 2;
      StopBits := 2;
      Baudrate := 2400;
    end else begin
      Baudrate := 9600;
    end;
    RtsFlowControl := FALSE;
    TimerInterval := 100;
    HexMonitor := not UseModem;
  end;
end;

procedure TfrmLogger.FormCreate(Sender: TObject);
begin
  Comm1 := TComm.Create(self);
  Comm2 := TComm.Create(self);
  InitComm(Comm1);
  InitComm(Comm2);
end;

procedure TfrmLogger.btnStartClick(Sender: TObject);
var
  Ok : boolean;
  d : TDateTime;
begin
  System.assign(log_file, Edit1.Text);
  if not FileExists(Edit1.Text) then
    System.Rewrite(log_file)
  else
    System.Append(log_file);
  writeln(log_file);
  d := now;
  writeln(log_file, 'Start: ', DateToStr(d), ' ', TimeToStr(d));

  Comm1.ComPortNr := cbxComm1.ItemIndex + 1;
  Ok := Comm1.InitCOM;
  Comm2.ComPortNr := cbxComm2.ItemIndex + 1;
  if Ok then begin
    Ok := Comm2.InitCOM;
    if not Ok then
      Comm1.Close;
  end;
  flush_delay := 100;
  written := 0;
  if Ok then begin
    Edit1.Enabled := false;
    Label1.Enabled := false;
    btnStart.Enabled := false;
    cbxComm1.Enabled := false;
    cbxComm2.Enabled := false;
    btnStart.Visible := false;
    btnStop.Visible := true;
    Timer1.Enabled := true;
  end;
end;

procedure TfrmLogger.btnStopClick(Sender: TObject);
var
  d : TDateTime;
begin
  try
    Timer1.Enabled := false;
    Edit1.Enabled := true;
    Label1.Enabled := true;
    btnStart.Enabled := true;
    cbxComm1.Enabled := true;
    cbxComm2.Enabled := true;
    btnStart.Visible := true;
    btnStop.Visible := false;

    writeln(log_file);
    d := now;
    writeln(log_file, 'Stop: ', DateToStr(d), ' ', TimeToStr(d));
    System.CloseFile(log_file);
  finally
    Comm1.ExitCOM;
    Comm2.ExitCOM;
  end;
end;

const digits : array [0..15] of char =
  ('0', '1', '2', '3', '4', '5', '6', '7',
   '8', '9', 'a', 'b', 'c', 'd', 'e', 'f');

procedure TfrmLogger.Timer1Timer(Sender: TObject);
var
  w : integer;
begin
  w := Comm1.ReadFromCOM;
  if w < $100 then begin
    Comm2.WriteToCOM(char(w));
    write(log_file, digits[w shr 4], digits[w and $f]);
    inc(written);
  end;
  w := Comm2.ReadFromCOM;
  if w < $100 then begin
    Comm1.WriteToCOM(char(w));
    write(log_file, '.', digits[w shr 4], digits[w and $f]);
    inc(written);
  end;
  dec(flush_delay);
  if (written > 100) or (flush_delay <= 0) then begin
    System.Flush(log_file);
    flush_delay := 1000;
    written := 0;
  end;

end;

end.
